import numpy as np

def nsphere_to_cartesian(r, angles):
    a = np.concatenate((np.array([2 * np.pi]), angles))
    si = np.sin(a)
    si[0] = 1
    si = np.cumprod(si)
    co = np.cos(a)
    co = np.roll(co, -1)
    return si * co * r

def distance(point1, point2):
    sum_pows = 0
    if len(point1) == len(point2):
        for i in range(0, len(point1)):
            sum_pows += pow(point2[i] - point1[i], 2)
        return np.sqrt(sum_pows)
    else:
        raise ValueError

def distance_squared(point1, point2):
    sum_pows = 0
    if len(point1) == len(point2):
        for i in range(0, len(point1)):
            sum_pows += pow(point2[i] - point1[i], 2)
        return sum_pows
    else:
        raise ValueError

def distance_taxicab(point1, point2):
    sum_abs = 0
    if len(point1) == len(point2):
        for i in range(0, len(point1)):
            sum_abs += abs(point2[i] - point1[i])
        return sum_abs
    else:
        raise ValueError


from problems.IProblem import IProblem

from numpy.random import RandomState

class Knapsack_Categorical(IProblem):
    def __init__(self, knapsack_capacity, total_posible_elements, seed, max_cost, max_value, elements=None):
        self.knapsack_capacity = knapsack_capacity
        if elements is not None and type(elements) == type([]):
            self.elements = elements
            self.total_posible_elements = len(elements)
        else:
            self.total_posible_elements = total_posible_elements
            rnd = RandomState(seed)
            #list of tuples with value and cost
            self.elements = []
            #fill elements
            for _ in range(total_posible_elements):
                self.elements.append((rnd.uniform(0, max_value), rnd.uniform(0, max_cost)))

    def get_possible_categories(self):
        #categories
        categories = ["is", "not"]
        categories_all_elements = [categories for _ in range(self.total_posible_elements)]
        return categories_all_elements

    def objective_function(self, point):
        if not self.is_valid(point):
            return False
        total_sum = 0
        for i in range(len(point)):
            if point[i] == "is":
                total_sum += self.elements[i][0]
        return total_sum

    def preprocess_function(self, point):
        return point

    def repair_function(self, point):
        i = -1
        times_founded = 0
        while not self.is_valid(point):
            found = False
            while not found:
                i += 1
                if i >= len(point):
                    print("que pedo wey, i es mayor que len en repair")
                    print("i value: ", i)
                    print("times founded: ", times_founded)
                    print("punto: ", point)
                    sum_elements_in, elements_value_in, elements_cost_in = self.get_values_in(point, True)
                    print(sum_elements_in)
                    print(elements_value_in)
                    print(elements_cost_in)
                    exit()
                if point[i] == "is":
                    point[i] = "not"
                    found = True
                    times_founded +=1
        return point    
   
    def get_values_in(self, point, verbose):
        sum_elements_in = 0
        elements_value_in = []
        elements_cost_in = []
        for i in range(len(point)):
            if point[i] == "is":
                sum_elements_in +=1
                elements_value_in.append(self.elements[i][0])
                elements_cost_in.append(self.elements[i][1])
        if verbose:
            print("sum: ", sum_elements_in)
            print("value total: ", sum(elements_value_in))
            print("cost total: ", sum(elements_cost_in))
            print("max knapsack capacity: ", self.knapsack_capacity)
            print("values: ", elements_value_in)
            print("costs: ", elements_cost_in)
        return sum_elements_in, elements_value_in, elements_cost_in

    def is_valid(self, point) -> bool:
        total_sum = 0
        for i in range(len(point)):
            if point[i] == "is":
                total_sum += self.elements[i][1]
        if total_sum > self.knapsack_capacity:
            return False
        return True


from anmetal.population.AFSA.AFSAMH_Real import AFSAMH_Real
from anmetal.population.SillyRandom.GreedyMH_Real import GreedyMH_Real
from anmetal.population.SillyRandom.GreedyMH_Real_WithLeap import GreedyMH_Real_WithLeap
from anmetal.population.PSO.PSOMH_Real import PSOMH_Real
from anmetal.population.PSO.PSOMH_Real_WithLeap import PSOMH_Real_WithLeap


#from problems.nphard_real.partition__and_subset_sum import Partition_Real, Subset_Real
import anmetal.problems.nonlinear_functions.two_inputs as problems_2

import matplotlib.pyplot as plt
from numpy.random import RandomState
import numpy as np
import os
from os.path import exists, join

import argparse
import shutil

parser = argparse.ArgumentParser(description='Metaheuristic image graph maker')
#parser.add_argument("--a", default=1, type=int, help="This is the 'a' variable")

parser.add_argument("--seed", default=0, type=int, help="The integer to be seed of random number generator")
parser.add_argument("--mh", default="", type=str, help="\
    Name of Metaheuristic, can be of of:\n\
    AFSA\n\
    PSO\n\
    PSOWL\n\
    Greed\n\
    GreedWL")
parser.add_argument("--problem", default="Goldsteinprice", type=str, help="\
    Name of the problem, can be one of:\
    Camelback\n\
    Goldsteinprice\n\
    Pshubert1\n\
    Pshubert2\n\
    Shubert\n\
    Quartic")
parser.add_argument("--verbose", default=1, type=int, help="1 if print logs, 0 if not print")
parser.add_argument("--iterations", default=100, type=int, help="Number of iterations in Metaheuristic")
parser.add_argument("--population", default=30, type=int, help="Number of solutions in Metaheuristic")

args = parser.parse_args()

print("args: ", args)

to_verbose : bool = True if args.verbose == 1 else False
seed: int = args.seed
random_generator = RandomState(seed)
iterations = args.iterations
population = args.population

probs_dict = {
    "Camelback": problems_2.Camelback,
    "Goldsteinprice": problems_2.Goldsteinprice,
    "Pshubert1": problems_2.Pshubert1,
    "Pshubert2": problems_2.Pshubert2,
    "Shubert": problems_2.Shubert,
    "Quartic": problems_2.Quartic
}

prob = probs_dict[args.problem] if args.problem in probs_dict else problems_2.Camelback
#partition_problem = Partition_Real(seed=seed, num_dims=200)

if str.lower(args.mh) == "afsa" or args.mh == "":
    mh = AFSAMH_Real(prob.get_limits()[0], prob.get_limits()[1], 2, False, prob.func, None, None)
    gen = mh.run_yielded(verbose=to_verbose, iterations=iterations, population=population, visual_distance_percentage=0.5, velocity_percentage=0.5, n_points_to_choose=3, crowded_percentage=0.7, its_stagnation=4, leap_percentage=0.3, stagnation_variation=0.4, seed=seed)
elif str.lower(args.mh) == "pso":
    mh = PSOMH_Real(prob.get_limits()[0], prob.get_limits()[1], 2, False, prob.func, None, None)
    gen = mh.run_yielded(verbose=to_verbose, iterations=iterations, population=population, omega=0.8, phi_g=1, phi_p=0.5)
elif str.lower(args.mh) == "psowl":
    mh = PSOMH_Real_WithLeap(prob.get_limits()[0], prob.get_limits()[1], 2, False, prob.func, None, None)
    gen = mh.run_yielded(verbose=to_verbose, iterations=iterations, population=population, omega=0.8, phi_g=1, phi_p=0.5)
elif str.lower(args.mh) == "greed":
    mh = GreedyMH_Real(prob.get_limits()[0], prob.get_limits()[1], 2, False, prob.func, None, None)
    gen = mh.run_yielded(verbose=to_verbose, iterations=iterations, population=population, stagnation_variation=0.4, its_stagnation=5, leap_percentage=0.8)
elif str.lower(args.mh) == "greedwl":
    mh = GreedyMH_Real(prob.get_limits()[0], prob.get_limits()[1], 2, False, prob.func, None, None)
    gen = mh.run_yielded(verbose=to_verbose, iterations=iterations, population=population, stagnation_variation=0.4, its_stagnation=5, leap_percentage=0.8)

folderpath = join("mh_graphs", args.mh+"_"+args.problem)
if exists(folderpath):
    shutil.rmtree(folderpath, ignore_errors=True)
os.makedirs(folderpath)



print("to start iterations")
colors_to_use = []
is_first : bool = True
for iteration, best_fitness_historical, best_bin_point, points_a, fts in gen:
    print("iteration: ", iteration)
    plt.xlim(prob.get_limits()[0], prob.get_limits()[1])
    plt.ylim(prob.get_limits()[0], prob.get_limits()[1])
    ps_ = np.copy(points_a)
    if is_first: #is_first para no depender del numero
        for i in range(len(ps_)):
            color = (random_generator.uniform(0.3, 1),
                    random_generator.uniform(0.3, 1),
                    random_generator.uniform(0.3, 1))
            colors_to_use.append(color)
            plt.plot(ps_[i][0], ps_[i][1], '*', color=color)
        last_points = np.copy(ps_) #copiar array
        is_first = False
    else:
        #line between points (2,5) -> (4,8)
        for i in range(len(ps_)):
            plt.plot([last_points[i][0], ps_[i][0]], [last_points[i][1], ps_[i][1]], '-', color=colors_to_use[i])
            plt.plot(ps_[i][0], ps_[i][1], '*', color=colors_to_use[i])
        last_points = np.copy(ps_) #copiar array
    plt.savefig(join(folderpath, "mhgraph_"+str(iteration)+".png"))
    plt.clf()


import anmetal.utils.points_utils as utils
from anmetal.population.IMetaheuristic import IMetaheuristic
from anmetal.population.PSO.Particle import SolutionParticle
import numpy as np
from numpy.random import RandomState
from typing import List, Callable

class PSOMH_Real(IMetaheuristic):
    def __init__(self, min_value: float, max_value: float, ndims: int, to_max: bool,
     objective_function: Callable[[List[float]], float],
      repair_function: Callable[[List[float]], List[float]],
      preprocess_function: Callable[[List[float]], List[float]] = None):
        self._group = []
        self._random_generator : RandomState = None
        self._min = min_value
        self._max = max_value
        self._ndims = ndims
        self._to_max = to_max

        self.objective_function = objective_function
        self.preprocess_function = \
         preprocess_function if preprocess_function is not None else lambda p: p
        self.repair_function = \
         repair_function if repair_function is not None else lambda p: p


    
    def run_yielded(self, iterations: int = 100, population: int =30, seed: int = None,
        omega: float=0.5, phi_p: float=1, phi_g: float=1, verbose:bool=False):
        self._iterations = iterations
        self._population = population
        self._seed = seed
        self._random_generator = RandomState(seed)

        self.omega = omega
        self.phi_p = phi_p
        self.phi_g = phi_g

        self.initialize_population(population)
        iteration = 1
        best_solution_historical = self.find_best_solution(self._group)
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        #yield
        points = [e.point for e in self._group]
        fts = [e.fitness for e in self._group]
        bin_point = self.preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
        while iteration <= iterations:
            if verbose:
                print("it: ", iteration, " fitness mejor: ", best_fitness_historical)
            for individual in self._group:
                #update the velocity
                for idim in range(len(individual.point)):
                    rp = self._random_generator.uniform()
                    rg = self._random_generator.uniform()
                    individual.velocity[idim] = omega * individual.velocity[idim] + \
                        phi_p*rp * (individual.best_point[idim] - individual.point[idim]) +\
                            phi_g*rg * (best_point_historical[idim] - individual.point[idim])
                result_point, fitness = self.Move(individual)
                individual.move_to(result_point, fitness)# self.objective_function(self.preprocess_function(result_point)))
                if self._to_max and individual.fitness > individual.best_fitness:
                    individual.set_best_point(result_point, fitness)
                if not self._to_max and individual.fitness < individual.best_fitness:
                    individual.set_best_point(result_point, fitness)
                # print("fitness del fish: ",fish.fish_id," es: ",fish.fitness)
            iteration += 1
            # print("seteando historicoooooooooooooooooooooooo")
            best_solution_it = self.find_best_solution(self._group)
            best_fitness_it = best_solution_it.fitness
            best_point_it = np.copy(best_solution_it.point)
            if self._to_max and best_fitness_it > best_fitness_historical:
                best_fitness_historical = best_fitness_it
                best_point_historical = best_point_it
            if not self._to_max and best_fitness_it < best_fitness_historical:
                best_fitness_historical = best_fitness_it
                best_point_historical = best_point_it
            #yield
            points = [e.point for e in self._group]
            fts = [e.fitness for e in self._group]
            bin_point = self.preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
        bin_point = self.preprocess_function(best_point_historical)
        # print("suma de 1s: ", np.sum(bin_point))
        #return best_fitness_historical, bin_point
        #yield
        points = [e.point for e in self._group]
        fts = [e.fitness for e in self._group]
        yield iteration, best_fitness_historical, bin_point, points, fts
    
    def run(self, iterations: int = 100, population: int =30, seed: int = None,
        omega: float=0.5, phi_p: float=1, phi_g: float=1, verbose:bool=False):
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, seed, omega, phi_p, phi_g, verbose):
            continue
        return best_fitness_historical, bin_point

    def initialize_population(self, population: int):
        for indivIndex in range(0, population):
            point, fitness = self.generate_random_point()
            velocity = self._random_generator.uniform(-(self._max - self._min),
                                                        (self._max - self._min),
                                                        self._ndims)
            individual = SolutionParticle(indivIndex, point, fitness, velocity)
            self._group.append(individual)

    def Move(self, individual): #get closer
        origin_point = np.copy(individual.point)
        for idim in range(len(origin_point)):
            origin_point[idim] += individual.velocity[idim]
        return self.repair_or_not(origin_point)

    def generate_random_point(self):
        cartesian_point = self._random_generator.uniform(self._min,
                                                        self._max,
                                                        self._ndims)
        return self.repair_or_not(cartesian_point)
    
    def repair_or_not(self, cartesian_point: List[float]):
        cartesian_point = IMetaheuristic.cut_mod_point(cartesian_point, self._min, self._max)
        fitness = self.objective_function(self.preprocess_function(cartesian_point))
        if not fitness:
            new_point = self.repair_function(cartesian_point)
            fitness = self.objective_function(self.preprocess_function(new_point))
            return new_point, fitness
        else:
            return cartesian_point, fitness
    
    


from anmetal.population.AFSA.AFSAMH_Real import AFSAMH_Real
from anmetal.population.SillyRandom.GreedyMH_Real import GreedyMH_Real
from anmetal.population.SillyRandom.GreedyMH_Real_WithLeap import GreedyMH_Real_WithLeap
from anmetal.population.PSO.PSOMH_Real import PSOMH_Real
from anmetal.population.PSO.PSOMH_Real_WithLeap import PSOMH_Real_WithLeap


from problems.nphard_real.partition__and_subset_sum import Partition_Real, Subset_Real

to_use = [
    "AFSA",
    "Greedy",
    "GreedyWL",
    "PSO",
    "PSOWL"
    ]

partition_problem = Partition_Real(seed=0, num_dims=200)
subset_problem = Subset_Real(seed=0, num_dims=200)


problem_to_solve = "partition sum"
#problem_to_solve = "subset sum"

to_verbose = True

if "AFSA" in to_use:
    print("create afsa")
    if problem_to_solve == "partition sum":
        mh = AFSAMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = AFSAMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run afsa")
    fit, pt = mh.run(verbose=to_verbose, visual_distance_percentage=0.5, velocity_percentage=0.5, n_points_to_choose=3, crowded_percentage=0.7, its_stagnation=4, leap_percentage=0.3, stagnation_variation=0.4, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "GreedyWL" in to_use:
    print("create GreedyWL")
    if problem_to_solve == "partition sum":
        mh = GreedyMH_Real_WithLeap(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = GreedyMH_Real_WithLeap(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run greedy")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, stagnation_variation=0.4, its_stagnation=5, leap_percentage=0.8, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "Greedy" in to_use:
    print("create Greedy")
    if problem_to_solve == "partition sum":
        mh = GreedyMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = GreedyMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run greedy")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "PSO" in to_use:
    print("create PSO")
    if problem_to_solve == "partition sum":
        mh = PSOMH_Real(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = PSOMH_Real(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run PSO")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, omega=0.8, phi_g=1, phi_p=0.5,seed=115)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)

if "PSOWL" in to_use:
    print("create PSOWL")
    if problem_to_solve == "partition sum":
        mh = PSOMH_Real_WithLeap(partition_problem.min_x, partition_problem.max_x, partition_problem.ndim, False, partition_problem.objective_function, partition_problem.repair_function, partition_problem.preprocess_function)
    if problem_to_solve == "subset sum":
        mh = PSOMH_Real_WithLeap(subset_problem.min_x, subset_problem.max_x, subset_problem.ndim, False, subset_problem.objective_function, subset_problem.repair_function, subset_problem.preprocess_function)
    print("to run PSO")
    fit, pt = mh.run(verbose=to_verbose, iterations=100, population=30, omega=0.8, phi_g=1, phi_p=0.5 ,seed=115, stagnation_variation=0.4, its_stagnation=5, leap_percentage=0.8)
    print(fit)
    print(pt)
    if problem_to_solve == "partition sum":
        partition_problem.print_difference_subsets(pt)
    if problem_to_solve == "subset sum":
        subset_problem.print_difference_subsets(pt)
    



from abc import ABCMeta, abstractmethod

class IProblem:
    __metaclass__ = ABCMeta

    @abstractmethod
    def objective_function(self, point):
        raise NotImplementedError

    @abstractmethod
    def preprocess_function(self, point):
        raise NotImplementedError

    @abstractmethod
    def repair_function(self, point):
        raise NotImplementedError
    


#functions taken from https://doi.org/10.1007/s00521-017-3088-3
#Applying modified cuckoo search algorithm for solving systems of nonlinear equations
import numpy as np

#all were tested in the paper with 20 dimensions
class Brown1:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for x in x_array:
            result += x-3
        result = result**2
        for i in range(0, len(x_array)-1):
            result += pow(10, -3)*((x_array[i]-3)**2) - (x_array[i]-x_array[i+1])+np.exp(20*(x_array[i]-x_array[i+1]))
        return result

    @staticmethod
    def get_limits():
        return [-1, 4]
    
    @staticmethod
    def get_theoretical_optimum():
        return 2.0
    
    @staticmethod
    def get_type():
        return "min"

class Brown3:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for i in range(0, len(x_array)-1):
            result += pow((x_array[i])**2, (x_array[i+1]**2)+1)
            result += pow((x_array[i+1])**2, (x_array[i]**2)+1)
        return result

    @staticmethod
    def get_limits():
        return [-1, 4]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"

class F10n:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for i in range(0, len(x_array)-1):
            result += (x_array[i]-1)**2 * (1 + 10*(np.sin(np.pi*x_array[i+1]))**2)
        result += 10*(np.sin(np.pi * x_array[0])**2)
        result += (x_array[len(x_array)-1] - 1)**2
        result = (np.pi/ len(x_array)) * result
        return result

    @staticmethod
    def get_limits():
        return [-10, 10]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"

class F15n:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for i in range(0, len(x_array)-1):
            result += (x_array[i]-1)**2 * (1 + (np.sin(3*np.pi*x_array[i+1]))**2)
        result += (np.sin(3*np.pi * x_array[0])**2)
        result += (1/10.0) * (x_array[len(x_array)-1] - 1)**2 * (1+ np.sin(2*np.pi*x_array[len(x_array)-1]))
        result = (1/10.0) * result
        return result

    @staticmethod
    def get_limits():
        return [-10, 10]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"



#functions taken from https://doi.org/10.1007/s00521-018-3512-3
#An Improved dynamic self-adaption cuckoo search algorithm based on collaboration between subpopulations

class Sphere:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for x in x_array:
            result += x**2
        return result

    @staticmethod
    def get_limits():
        return [-100, 100]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"

class Rosenbrock:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for i in range(0, len(x_array)-1):
            result += (100.0*(x_array[i+1] - (x_array[i]**2))**2) + ((x_array[i] - 1)**2)
        return result

    @staticmethod
    def get_limits():
        return [-30, 30]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"


class Griewank:
    @staticmethod
    def func(x_array: list) -> float:
        r_sum = 0.0
        for x in x_array:
            r_sum += x**2
        r_prod = 1.0
        for i in range(len(x_array)):
            r_prod *= np.cos(x_array[i]/float(np.sqrt(i+1)))
        result = (1/4000.0)*r_sum - r_prod + 1.0
        return result

    @staticmethod
    def get_limits():
        return [-600, 600]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"


class Rastrigrin:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for x in x_array:
            result += x**2 - 10*np.cos(2 * np.pi * x) +10
        return result

    @staticmethod
    def get_limits():
        return [- 5.12, 5.12]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"


class Sumsquares:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for i in range(len(x_array)):
            result += (i+1)*(x_array[i]**2)
        return result

    @staticmethod
    def get_limits():
        return [-10, 10] 
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"


class Michalewicz:
    @staticmethod
    def func(x_array: list, m:float = 10) -> float:
        result = 0.0
        for i in range(len(x_array)):
            result += np.sin(x_array[i]) * ((np.sin( ((i+1)*(x_array[i]**2))/ np.pi ))**(2*m))
        result = -result
        return result

    @staticmethod
    def get_limits():
        return [0, np.pi]
    
    @staticmethod
    def get_theoretical_optimum():
        #-1.8013 if d=2
        #-4.687658 if d=5
        #-9.66015 if d=10
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"


class Quartic:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for i in range(len(x_array)):
            result += (i+1)*(x_array[i]**4) + np.random.random()
        return result

    @staticmethod
    def get_limits():
        return [- 1.28, 1.28]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"


class Schwefel:
    @staticmethod
    def func(x_array: list) -> float:
        result = 0.0
        for x in x_array:
            result += x * np.sin(np.sqrt(np.abs(x)))
        result = ( 418.9829 * len(x_array) ) - result
        return result

    @staticmethod
    def get_limits():
        return [-500, 500]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0#- 418.9 * 5.0 #- 418.9 X 5
    
    @staticmethod
    def get_type():
        return "min"


class Penalty:
    @staticmethod
    def func(x_array: list) -> float:
        def y_i(x_i):
            return 1.0 + ( (x_i+1.0)/4.0 )
        def u(x_i, a, k, m):
            if x_i > a:
                return k * ((x_i - a)**m)
            elif -a <= x_i <= a:
                return 0
            else: # x_i < -a
                return k * ((-x_i - a)**m)
        sum_u = 0.0
        for x in x_array:
            sum_u += u(x, 10, 100, 4)
        result = 0.0
        for i in range(len(x_array)-1):
            result += (y_i(x_array[i]) - 1)**2 * ( 1 + 10*np.sin(np.pi * y_i(x_array[i+1]))**2 )
        result += 10 * np.sin(np.pi*y_i(x_array[0]))
        result += (y_i(x_array[len(x_array)-1]) + 1)**2
        result = (np.pi / float(len(x_array))) * result
        result += sum_u
        return result

    @staticmethod
    def get_limits():
        return [-50, 50]
    
    @staticmethod
    def get_theoretical_optimum():
        return 0.0
    
    @staticmethod
    def get_type():
        return "min"



import math
import anmetal.iterative_methods as it_optimization


##example of newton method for 100^100 in video of "derivando" channel
class x_times_x:
    def __init__(self, result=0):
        self.result = result
    def __call__(self, x):
        return pow(x, x) - self.result

#def x_times_x(x, result=0):
#    return pow(x, x) - result

def x_times_x_derivative(x):
    return pow(x, x) * (1 + math.log(x))

x_times_x100_func = x_times_x(100)
x_times_x100_res = it_optimization.newton_method(x_times_x100_func, x_times_x_derivative, 3.5, 20)
print("x^x = 100, then x is aprox:",x_times_x100_res, " what gives: ", pow(x_times_x100_res, x_times_x100_res))


##example of euler method in wikipedia

def dy_dx(x, y):
    return math.sin(x) - math.log(y)

#exact value es y=0.3325459
y_res = it_optimization.euler_method(dy_dx, 0.13, 0.14, 0.32, 4)
print("euler example: ", y_res)

def dy_dx_line(x, y):
    #y = mx+b
    #dy = m
    return 2

y_res = it_optimization.euler_method(dy_dx_line, 6, 8, 25, 4) #y = x*2+13, 25=6*2+13
print("euler example 2 (29=8*2+13): ", y_res) # 29


import anmetal.problems.nonlinear_functions.n_inputs as problems_n
import anmetal.problems.nonlinear_functions.one_input as problems_1
import anmetal.problems.nonlinear_functions.two_inputs as problems_2

import numpy as np

#random values

def make_random_vector(cls_type, n):
    limits = cls_type.get_limits()
    vect = []
    if type(limits[0]) == type(list()):
        for i in range(len(limits)):
            vect.append(np.random.uniform(limits[i][0], limits[i][1]))
    elif type(limits[0]) == type(float()) or type(limits[0]) == type(int()):
        for i in range(n):
            vect.append(np.random.uniform(limits[0], limits[1]))
    else:
        raise ValueError("limits are no valid")
    return vect

cls_ns = [problems_n.Sumsquares, problems_n.Sphere, problems_n.Schwefel,
problems_n.Rosenbrock, problems_n.Rastrigrin, problems_n.Quartic, problems_n.Penalty,
#problems_n.Michalewicz, #the theorical min is not fixed, depends on dimension
problems_n.Griewank,
#below from first paper
problems_n.F15n, problems_n.F10n, problems_n.Brown3, problems_n.Brown1]


cls_1s = [problems_1.F1, problems_1.F3]
cls_2s = [problems_2.Camelback, problems_2.Goldsteinprice]

def eval_random(cls_list, n):
    rs = {}
    for cls_type in cls_list:
        vect = make_random_vector(cls_type, n)
        val = cls_type.func(vect)
        print(str(cls_type))
        print("value obtained: ", val, " and theoretical min: ", cls_type.get_theoretical_optimum())
        print("difference: ", abs(cls_type.get_theoretical_optimum()-val))
        rs[str(cls_type)] = {"vect": vect, "val": val, "theory": cls_type.get_theoretical_optimum()}
        if val < cls_type.get_theoretical_optimum():
            print("!"*40)
            print("WTF, it's less than minimum, something wrong")
    return rs

#random points
if False:
    print("#"*80)
    print("ns with 20 length vector")
    rn = eval_random(cls_ns, 30)
    print("#"*80)
    print("1s")
    r1 = eval_random(cls_1s, 1)
    print("#"*80)
    print("2s")
    r2 = eval_random(cls_2s, 2)


from population.AFSA.AFSAMH_Real import AFSAMH_Real
from population.SillyRandom.GreedyMH_Real import GreedyMH_Real
from population.SillyRandom.GreedyMH_Real_WithLeap import GreedyMH_Real_WithLeap
from population.PSO.PSOMH_Real import PSOMH_Real
from population.PSO.PSOMH_Real_WithLeap import PSOMH_Real_WithLeap

def eval_mh(cls_list, n, mh_to_use, verbose=False):
    rs = {}
    for cls_type in cls_list:
        limits = cls_type.get_limits()
        if type(limits[0]) == type(float()) or type(limits[0]) == type(int()):
            min_x = limits[0]
            max_x = limits[1]
        else:
            print("skipping ", str(cls_type))
            continue
        if mh_to_use == "AFSA":
            mh = AFSAMH_Real(min_x, max_x, n, False, cls_type.func, None, None)
            fit, pt = mh.run(verbose=verbose, visual_distance_percentage=0.5, velocity_percentage=0.5, n_points_to_choose=3, crowded_percentage=0.7, its_stagnation=4, leap_percentage=0.3, stagnation_variation=0.4, seed=115)
        if mh_to_use == "Greedy":
            mh = GreedyMH_Real(min_x, max_x, n, False, cls_type.func, None, None)
            fit, pt = mh.run(verbose=verbose, iterations=100, population=30, seed=115)
        if mh_to_use == "GreedyWL":
            mh = GreedyMH_Real_WithLeap(min_x, max_x, n, False, cls_type.func, None, None)
            fit, pt = mh.run(verbose=verbose, iterations=100, population=30, stagnation_variation=0.4, its_stagnation=5, leap_percentage=0.8, seed=115)
        if mh_to_use == "PSO":
            mh = PSOMH_Real(min_x, max_x, n, False, cls_type.func, None, None)
            fit, pt = mh.run(verbose=verbose, iterations=100, population=30, omega=0.8, phi_g=0.5, phi_p=0.5 ,seed=115)
        if mh_to_use == "PSOWL":
            mh = PSOMH_Real_WithLeap(min_x, max_x, n, False, cls_type.func, None, None)
            fit, pt = mh.run(verbose=verbose, iterations=100, population=30, omega=0.8, phi_g=0.5, phi_p=0.5 ,seed=115, stagnation_variation=0.4, its_stagnation=5, leap_percentage=0.8)
        #vect = make_random_vector(cls_type, n)
        #val = cls_type.func(vect)
        print(str(cls_type))
        print("value obtained: ", fit, " and theoretical min: ", cls_type.get_theoretical_optimum())
        print("difference: ", abs(cls_type.get_theoretical_optimum()-fit))
        rs[str(cls_type)] = {"vec": pt, "val": fit, "theory": cls_type.get_theoretical_optimum(),
        "diff": abs(cls_type.get_theoretical_optimum()-fit)}
        if fit < cls_type.get_theoretical_optimum():
            print("!"*40)
            print("WTF, it's less than minimum, something wrong")
    return rs

if True:
    print("#"*80)
    print("ns with 30 length vector")
    mh_to_use = "AFSA"  # can be AFSA, Greedy, GreedyWL, PSO, PSOWL
    mh_rn = eval_mh(cls_ns, 30, mh_to_use)
    print("#"*80)
    print("1s")  # problems with 1D vector
    mh_r1 = eval_mh(cls_1s, 1, mh_to_use)
    print("#"*80)
    print("2s")  # problems with 2D vector
    mh_r2 = eval_mh(cls_2s, 2, mh_to_use)



import numpy as np

class Sudoku:
    def __init__(self, initial_state: list=None):
        self.state = initial_state
        if self.state is None:
            self.state = np.zeros((9,9))
        self.categories = [1,2,3,4,5,6,7,8,9]

    def set_categories(self, categories):
        self.categories = categories

    @staticmethod
    def is_row_violating(row, comming_from: str = ""):
        if len(row) != 9:
            ValueError("state row, column or box should be of size 9. comming from: "+comming_from)
        ncats = 0
        violated_row = False
        for x in row:
            if x not in self.categories:
                violated_row = True
                break
            equals_filter = row == x
            if np.sum(equals_filter) > 1:
                violated_row = True
                break
            ncats +=1
        if violated_row or ncats != 9:
            return True
        return False

    def get_violations(self):
        if type(self.state) != type(np.zeros(1)):
            raise ValueError("state should be numpy ndarray")
        n_violations = 0
        #horizontal
        for ihoriz in range(1,10):
            if Sudoku.is_row_violating(self.state[ihoriz], "horizontal: "+str(ihoriz)):
                n_violations += 1
        #vertical
        for ivert in range(1,10):
            if Sudoku.is_row_violating(self.state[:,ivert], "vertical: "+str(ivert)):
                n_violations += 1
        #boxes
        #self.states
        #[0:3,0:3]  [0:3,3:6]   [0:3,6:9]
        #[3:6,0:3]  [3:6,0:3]   [3:6,6:9]
        #[6:9,0:3]  [6:9,0:3]   [6:9,6:9]
        for ih1 in [0,3,6]:
            ih2 = ih1+3
            for ivert1 in [0,3,6]:
                ivert2 = ivert1 + 3
                box = self.state[ih1:ih2,ivert1:ivert2]
                if Sudoku.is_row_violating(box.flatten(), "box ["+str(ih1)+":"+str(ih2)+","+str(ivert1)+":"+str(ivert2)+"]"):
                    n_violations +=1



from anmetal.population.IMetaheuristic import IMetaheuristic
from anmetal.population.ISolution import SolutionWithId
import numpy as np
from numpy.random import RandomState
from typing import List, Callable, Tuple

class GeneticMH_Categorical(IMetaheuristic):
    def __init__(self, categorics: list, ndims: int, to_max: bool,
     objective_function: Callable[[list], float],
      repair_function: Callable[[list], list],
      preprocess_function: Callable[[list], list] = None):
        self._group: List[SolutionWithId] = []
        self._random_generator: RandomState = None
        
        self._categorics: list = categorics
        self._ndims: int = ndims
        self._to_max: bool = to_max

        self.objective_function = objective_function
        self.preprocess_function = \
         preprocess_function if preprocess_function is not None else lambda p: p
        self.repair_function = \
         repair_function if repair_function is not None else lambda p: p
    
    def run_yielded(self, iterations: int = 100, population: int =30,
        elitist_percentage: float = 0.3, mutability: float = 0.1, fidelity: bool = True,
        mutation_in_parents: bool = True, seed: int = None, verbose: bool = False):
        self._iterations = iterations
        self._population = population
        self._seed: int = seed

        self._mutability = mutability
        self._fidelity = fidelity
        self._mutation_in_parents = mutation_in_parents
        self.movements = {
            "random": 0,
            "init": 0,
            "select_elite": 0,
            "recombination": 0,
            "mutation": 0,
            "crossing": 0
        }
        
        self._random_generator: RandomState = RandomState(seed)

        self.initialize_population(population)
        iteration = 1
        best_solution_historical = self.find_best_solution(self._group)
        best_fitness_historical = best_solution_historical.fitness
        best_point_historical = np.copy(best_solution_historical.point)
        #yield
        points = [e.point for e in self._group]
        fts = [e.fitness for e in self._group]
        bin_point = self.preprocess_function(best_point_historical)
        yield iteration, best_fitness_historical, bin_point, points, fts
        while iteration <= iterations:
            if verbose:
                print("it: ", iteration, " best historical fitness: ", best_fitness_historical)
            elite = self.select_elite(elitist_percentage)
            self.recombination(elite)
            if not self._mutation_in_parents:
                for individual in self._group:
                    individual = self.mutate_individual(individual.point)
            iteration += 1
            # print("setting historical")
            best_solution_it = self.find_best_solution(self._group)
            best_fitness_it = best_solution_it.fitness
            best_point_it = np.copy(best_solution_it.point)
            if self._to_max and best_fitness_it > best_fitness_historical:
                best_fitness_historical = best_fitness_it
                best_point_historical = best_point_it
            if not self._to_max and best_fitness_it < best_fitness_historical:
                best_fitness_historical = best_fitness_it
                best_point_historical = best_point_it
            #yield
            points = [e.point for e in self._group]
            fts = [e.fitness for e in self._group]
            bin_point = self.preprocess_function(best_point_historical)
            yield iteration, best_fitness_historical, bin_point, points, fts
        bin_point = self.preprocess_function(best_point_historical)
        # print("suma de 1s: ", np.sum(bin_point))
        #return best_fitness_historical, bin_point
        #yield
        points = [e.point for e in self._group]
        fts = [e.fitness for e in self._group]
        yield iteration, best_fitness_historical, bin_point, points, fts
    
    def run(self, iterations: int = 100, population: int =30,
        elitist_percentage: float = 0.3, mutability: float = 0.1, fidelity: bool = True,
        mutation_in_parents: bool = True, seed: int = None, verbose: bool = False):
        for _, best_fitness_historical, bin_point, _, _ in self.run_yielded(
            iterations, population, elitist_percentage, mutability, 
            fidelity, mutation_in_parents, seed, verbose):
            continue
        return best_fitness_historical, bin_point

    def initialize_population(self, population: int):
        self.movements["init"] += 1
        self._group: List[SolutionWithId] = []
        for indivIndex in range(0, population):
            point, fitness = self.generate_random_point()
            individual = SolutionWithId(indivIndex, point, fitness)
            self._group.append(individual)

    def sort_group(self):
        fitnesses: List[float] = []
        for individual in self._group:
            fitnesses.append(individual.fitness)
        newgroup:List[SolutionWithId] = []
        for sorted_index in np.argsort(fitnesses):
            newgroup.append(self._group[sorted_index])
        self._group:List[SolutionWithId] = newgroup

    def select_elite(self, elitist_percentage: float) -> List[SolutionWithId]:
        self.movements["select_elite"] +=1
        self.sort_group()
        if int(len(self._group)*elitist_percentage) <= 0:
            raise Exception("Elitist percentage is too low")
        return self._group[:int(len(self._group)*elitist_percentage)]
    
    def recombination(self, elite: List[SolutionWithId]):
        self.movements["recombination"] += 1
        """
        I have an elite, subset of universe, cross it to complete the new universe
        without fidelity:
        - I create an array of indices [0...len(elite)]
        - I shuffle it (permutation)
        - I make a for 0 to population
        - for each individual, i search a couple and i make children
        with fidelity:
        - I create an array of indices [0...len(elite)]
        - I shuffle it (permutation)
        - I split that array int 2 (male and female, just taking the half)
        - I make a couple for each of them, coupling the indices (having in account if that individual is single)
        - I make a for 0 to population and I make children, using the same order in the shuffling and repeating
        """
        newgroup:List[SolutionWithId] = []
        indexes:List[int] = [i for i in range(len(elite))]
        self._random_generator.shuffle(indexes)
        if self._fidelity:
            indexes_len:int = int(len(indexes)/2.0)
            males:List[int] = indexes[:indexes_len]
            females:List[int] = indexes[indexes_len:indexes_len+len(males)]
        i_parent:int = 0
        for i_individual in range(self._population):
            if self._fidelity:
                point, fitness = self.cross_couple(elite[males[i_parent]], elite[females[i_parent]])
                individual = SolutionWithId(i_individual, point, fitness)
                newgroup.append(individual)
                i_parent += 1
                if i_parent >= len(males):
                    i_parent = 0
            else: #not fidelity
                #search a couple for elite[i_parent]
                couple_index:int = self._random_generator.randint(0, len(indexes)-1)
                #to avoid cross with itself
                couple_index:int = couple_index if couple_index >= i_parent else couple_index+1
                point, fitness = self.cross_couple(elite[indexes[i_parent]], elite[indexes[i_parent]])
                individual = SolutionWithId(i_individual, point, fitness)
                newgroup.append(individual)
                i_parent +=1
                if i_parent >= len(indexes):
                    i_parent = 0
        self._group:List[SolutionWithId] = newgroup


    def cross_couple(self, individual1: SolutionWithId, individual2: SolutionWithId) -> Tuple[list, float]:
        self.movements["crossing"] += 1
        individual_point1 = individual1.point
        individual_point2 = individual2.point
        if len(individual_point1) != len(individual_point2):
            raise Exception("length of points are different when crossing couples")
        if self._mutation_in_parents:
            individual_point1:list = self.mutate_individual(individual1).point
            individual_point2:list = self.mutate_individual(individual2).point
        indexes = [i%2 for i in  range(len(individual_point1))]
        self._random_generator.shuffle(indexes)
        result = []
        for iIndex in range(len(indexes)):
            if indexes[iIndex] == 0:
                result.append(individual_point1[iIndex])
            else:
                result.append(individual_point2[iIndex])
        return self.repair_or_not(result)


    def mutate_individual(self, individual: SolutionWithId):
        self.movements["mutation"] += 1
        individual_point = individual.point
        indexes_to_shuffle = [i for i in  range(len(individual_point))]
        self._random_generator.shuffle(indexes_to_shuffle)
        indexes_to_shuffle = indexes_to_shuffle[:int(len(indexes_to_shuffle)*self._mutability)]
        for index_to_shuffle in indexes_to_shuffle:
            individual_point[index_to_shuffle] = \
                 self._categorics[index_to_shuffle]\
                     [self._random_generator.randint(0, len(self._categorics[index_to_shuffle]))]
        point, fitness = self.repair_or_not(individual.point)
        return SolutionWithId(individual.get_id(), point, fitness)

    def generate_random_point(self):
        point = [self._categorics[i][self._random_generator.randint(0, len(self._categorics[i]))]\
                 for i in range(0, self._ndims)]
        return self.repair_or_not(point)
    
    def repair_or_not(self, point: list) -> Tuple[list, float]:
        fitness = self.objective_function(self.preprocess_function(point))
        if not fitness:
            new_point = self.repair_function(point)
            fitness = self.objective_function(self.preprocess_function(new_point))
            return new_point, fitness
        else:
            return point, fitness




def euler_method(dy_times_dx_func, x_start, x_end, y0, n_steps):
    h = float(x_end - x_start) / float(n_steps)
    x = x_start
    y = y0
    for _ in range(n_steps):
        y = y + h * dy_times_dx_func(x, y)
        x = x + h
    return y

def newton_method(func, derivative_func, x_start, n_iterations, m=1):
    x = x_start
    for _ in range(n_iterations):
        x = x - m*(float(func(x)) / float(derivative_func(x)))
    return x


from anmetal.population.Genetic.GeneticMH_Categorical import GeneticMH_Categorical
from anmetal.population.Genetic.GeneticMH_Categorical_WithLeap import GeneticMH_Categorical_WithLeap

from anmetal.problems.nphard_categorical.knapsack import Knapsack_Categorical

#import numpy as np
from numpy.random import RandomState
from typing import List, Callable, Tuple

print("hello it is categorical mh test")
#general parameters
knapsack_capacity:float = 50.8
total_posible_elements:int = 50

knpsk = Knapsack_Categorical(knapsack_capacity, total_posible_elements, seed=0, max_value=5, max_cost=6)

print("Genetic")
print("#"*40)
print("#"*40)
mh = GeneticMH_Categorical(categorics=knpsk.get_possible_categories(), ndims=knpsk.total_posible_elements, to_max=True, objective_function=knpsk.objective_function, repair_function=knpsk.repair_function)
fit, pt = mh.run(verbose=True)
print("fitness:")
print(fit)
print("point:")
print(pt)
print("movements:")
print(mh.movements)

knpsk.get_values_in(pt, True)

print("#"*40)
print("#"*40)
print("Genetic with leap")
mh = GeneticMH_Categorical_WithLeap(categorics=knpsk.get_possible_categories(), ndims=knpsk.total_posible_elements, to_max=True, objective_function=knpsk.objective_function, repair_function=knpsk.repair_function)
fit, pt = mh.run(verbose=True)
print("fitness:")
print(fit)
print("point:")
print(pt)
print("movements:")
print(mh.movements)

knpsk.get_values_in(pt, True)